# -*-perl-*-
#
# htmlop.pl versin 0.1.1: Do operations on html documents.
#
# Original source from Bjrn Borud, without it I would not have atempted
# this. In this incarnation it bears no resemblance ot Bjrns code however.
# - Nicolai Langfeldt 18/11/95.
# 
# htmlop.pl does operations on html files, possebly many at a time:
# Operations:
# - Absolitify urls
# - Relativify urls
# - Gather list of urls
#
# Authors:
# - Nicolai Langfelft (janl@ifi.uio.no)

package htmlop;

require "wwwurl.pl";

# HTML operation codes for process_html.  The first argument is the html 
# document to do operations on.  It will not be changed.

$ABS = 1;		# Absolutify urls. Arg: Origin, Ret: doc
$IABS = 2;		# Internal abs. Arg: Origin
$REL = 3;		# Relativize all urls. Arg: Origin, Ret: doc
$IREL = 4;		# Internal rel. Arg: Origin
$LIST = 5;		# List all urls, Ret: urls
$CANON = 6;		# See that missing opening (and ending) SGML tags
			# are injected. Ret: doc

# process_html returns a array.  The first component of the array is the
# new html document. The rest of the array is the urls.  If a document is
# not to be returned a empty string is returned. If a url list is not
# to be returned a empty array is returned.

# HERE BE DRAGONS

# Where to find urls in various tags.  The second compoent is a comma
# separated list.
%urls = ('A','HREF',
	 'IMG','SRC,LOWSRC',
	 'FRAME','SRC',
	 'BODY','BACKGROUND');

sub gettoken {
    # Get one token from the argument, removing it from the argument.
    # BUG: There should be whitespace at the end of the examined string.
    local($c,$token);
    
    # Skip whitespace and newlines
    $_[$[] =~ s/^[\n\s]*//;

    return '' if ($_[$[] eq '');

    $c = substr($_[$[],$[,1);
    substr($_[$[],$[,1)='';

    if ($c eq '"') {	# Quoted material
	($token,$rest)=split('"',$_[$[],2);
	$_[$[]=$rest;
    } elsif ($c eq '=') {
	$token='=';
    } else {		# Non quoted material
	$_[$[] =~ m/[=\s\n]/;
	$_[$[] = $&.$';
	$token=$c.$`;
    }
#    print "Token: '$token'\t\tRest: '",$_[$[],"'\n";
    return $token;
}


sub tagtoken {
    # Pick the tag contents to pieces (also knonw as tokens) Return an
    # associative array of tag contents.  The tag keys are changed to
    # uppercase.  The tag values are left as is.
    local($tok,$lasttok,%tokens);
    # Append a space, gettoken needs it
    $_[$[].=' ';
    $lasttok='';
    while (1) {
	$tok=&gettoken($_[$[]);
	last if ($tok eq '');
	$tok =~ tr/a-z/A-Z/;
	if ($tok eq '=') {
	    warn "bad html\n" if ($lasttok eq '');
	    $tokens{$lasttok}.=&gettoken($_[$[]);
#	    print STDERR "STORED: $lasttok = ",$tokens{$lasttok},"\n";
	} else {
	    $tokens{$tok}='';
	    $lasttok=$tok;
	}
    }
    return %tokens;
}


sub gettag {
    # Pick out the following things from the remaining html doc:
    # Everything leading up to the first tag.  The first tag type
    # and the first tags contents.
    # Modify @_ directly to reduce number of copies of possebly huge
    # documents kept in memory at once.
    # Return the body, the tag name, and the tag contents (associative array)
    local(%tokens,$tagn,$tagc,$body,$tag);

    unless ($_[$[] =~ s/([^<]*)\<([^>]*)>//) {
	$body=$_[$[];
	$_[$[]='';
	return ($body,'',());
    }
    $body=$1;
    $tag=$2;

#    print "BODY: $body\n";
#    print "HELE TAG: $tag\n";

    # Examine tag contents
    ($tagn,$tagc) = split(/[\s\n]+/,$tag,2);
    $tagn="\U$tagn";
    return ($body,$tagn,()) if !defined($tagc);
    return ($body,$tagn,($tagc,'')) if (substr($tagn,$[,1) eq '!');
#    print "CONTENTS: $tagc\n";
    %tokens=();
    %tokens=&tagtoken($tagc) if defined($tagc);
    return ($body,$tagn,%tokens);
}


# This is meant for general consumption:

sub process {
    # Process a html file.  Into one end you put a html file.  Out
    # of the other end you get something dependent on the operations
    # you specified.
    # I cannot gobble my arguments. I need to examine them several times.

    local($arg,$i,$retdoc,$retlist,$canon,$newdoc,@urllist,$Q,$cont,$SGML);
    
    $retdoc=$canon=0;

    $i=$[+1;
    # Argument checking
    while (defined($_[$i])) {
	$arg=$_[$i];
	if ($arg == $ABS || $arg == $REL ) {
	    $retdoc=1;
	    $i++;
	} elsif ($arg == $IABS || $arg == $IREL ) {
	    $i++;
	} elsif ($arg == $LIST) { 
	    # no flags
	} elsif ($arg == $CANON) {
	    $retdoc=1;
	    $canon=1;
	} else {
	    die "htmlop: Incorrect invocation of html_process\n";
	}
	$i++;
    }
    
    $endhtml=0;
    $SGML=0;	# 0: No SGML opening tags, 1: !SGML seen, 2: !DOCTYPE seen
    		# -1: it's OK
    $newdoc='';
    $doc=shift(@_);
    while ($doc ne '') {
	($textpart,$tagname,%tagcont)=&gettag($doc);
#	print "Tekst: $textpart\n";
#	if ($tagname) {
#	    print "Tag: $tagname\n";
#	    foreach $key (keys %tagcont) {
#		print "Tagcont: $key=",$tagcont{$key},"\n";
#	    }
#	}
#	print "Doc: $doc\n";

	# Process the tag
	while (1) {	# Need a way to get out of this
	    # 'Canonize'
	    if ($canon) {
		$endhtml=1 if ($tagname eq '/HTML');
		if ($SGML!=-1 && $canon && $tagname ne '!--' ) {
		    $wholetag=$tagname;
		    $moretag=join('',keys %tagcont);
		    $wholetag.=' '.$moretag if ($moretag ne '');
		    if ($SGML==0) {
			$SGML=1;
			if ($tagname eq '!SGML') {
			    $tagname=$wholetag;
			    %tagcont=();
			    last;
			}
			$newdoc.='<!SGML "ISO 8879:1986">'."\n";
		    }
		    if ($SGML==1) {
			$SGML=2;
			if ($tagname eq '!DOCTYPE') {
			    $tagname=$wholetag;
			    %tagcont=();
			    last;
			}
			# We have no idea what DTD this doc follows 
			# so we put in something kinda non-comitting
			$newdoc.='<!DOCTYPE HTML PUBLIC "html.dtd">'."\n";
		    }
		    if ($SGML==2) {
			$SGML=-1;
			if ($tagname eq 'HTML') {
			    $tagname=$wholetag;
			    %tagcont=();
			    last;
			}
			$newdoc.='<HTML>'."\n";
		    }
		} # sgml!=-1
	    } #canon
	    
	    $i=$[-1;
	    while (defined($_[$i+1])) {
		$i++;
		$arg=$_[$i];
		if (! ($arg =~ /^\d+$/)) {
		    print STDERR "Args: ",join(',',@_),"\n";
		    print STDERR "ARG: $arg is not a opcode.\n";
		    exit(1);
		}
		if ($arg == $ABS || $arg == $IABS ) {
		    $i++;
		    next unless defined($urls{$tagname});
		    $origin=$_[$i];
		    foreach $tagkey (split(',',$urls{$tagname})) {
			$tagcont{$tagkey}=
			    &wwwurl'absolute($origin,$tagcont{$tagkey}) #'
				if defined($tagcont{$tagkey});
		    }
		} elsif ($arg == $REL || $arg == $IREL ) {
		    $i++;
		    next unless defined($urls{$tagname});
		    $origin=$_[$i];
		    foreach $tagkey (split(',',$urls{$tagname})) {
			$tagcont{$tagkey} =~ s/^$origin//
			    if defined($tagcont{$tagkey});
		    }
		} elsif ($arg == $LIST) {
		    foreach $tagkey (split(',',$urls{$tagname})) {
			push(@urllist,$tagcont{$tagkey}) 
			    if defined($tagcont{$tagkey});
		    }
		}
	    }
	    last;
	}

	# Assemble new doc if wanted
	if ($retdoc) {
	    $newdoc.=$textpart;
	    if ($tagname) {
		$newdoc.='<'.$tagname;
		foreach $key (keys %tagcont) {
		    $newdoc.=' '.$key;
		    if ($cont=$tagcont{$key}) {
			$Q='';
			$Q='"' if ($cont =~ m/[^\w\d]/);
			$newdoc.='='.$Q.$cont.$Q;
		    }
		}
		$newdoc.='>';
	    }
	}
#	print "NEW: $newdoc\n";
    }
    $newdoc.="</HTML>\n" if ($canon && !$endhtml);
    return ($newdoc,@urllist);
}

1;

__END__

# Test stuff

$doc = <<EOHTML;
<HTML>
<HEAD>
<TITLE>eyeNET</TITLE></HEAD>
<BODY BGCOLOR="#000000" TEXT="#ffffbb" LINK="#ffff00" VLINK="#aaaa66" ALINK="#ff0000">
<CENTER>
<A HREF="/cgi-bin/imagemap/eyenet"><IMG BORDER=0 ALT="[ interactive image ]" SRC="eyenet.gif" ISMAP></A>
<P>
<IMG SRC="http:/eye/bar.jpg">
<P>
<I>Annnnd...</I><BR>
<A HREF="eyenet1.htm">A link for the graphically impaired</A>
<P>

</CENTER>
</BODY>
</HTML>
EOHTML

# '

($newdoc,@URLS)=
    &process($doc,$CANON,
		  $ABS,'http://foobar:500/',$LIST);

# ($newdoc,@URLS)=
#    &process_html('<!DOCTYPE --><!-- leave alone><html>'.
#		  'Dete er litt tekst '.
#                  '<a href=foo name=#bar>Test</a> mer.</HTML>',
#		  $CANON,$ABS,'',$LIST);

print "DOC: ",$newdoc;
print "URLS: ",join(',',@URLS),"\n";

0;
